package com.gigaspaces.ucs.deploy;

import java.util.concurrent.TimeUnit;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.openspaces.admin.Admin;
import org.openspaces.admin.AdminFactory;
import org.openspaces.admin.esm.ElasticServiceManager;
import org.openspaces.admin.esm.deployment.ElasticDataGridDeployment;
import org.openspaces.admin.esm.deployment.MemorySla;
import org.openspaces.admin.gsa.GridServiceAgent;
import org.openspaces.admin.gsa.GridServiceContainerOptions;
import org.openspaces.admin.gsa.GridServiceManagerOptions;
import org.openspaces.admin.gsc.GridServiceContainer;
import org.openspaces.admin.gsm.GridServiceManager;
import org.openspaces.admin.pu.ProcessingUnit;
import org.openspaces.admin.pu.ProcessingUnitDeployment;
import org.openspaces.admin.space.Space;
import org.openspaces.grid.esm.ElasticScaleHandlerConfig;

public class Deploy {

	final static Logger logger = Logger.getLogger("UCS.Deploy");
	final static Level loggerLevel = Level.INFO;

	public static void main(final String[] args) throws InterruptedException {

		logger.setLevel(loggerLevel);
		// Handler consoleHandler = new ConsoleHandler();
		// consoleHandler.setLevel(loggerLevel);
		// logger.addHandler(consoleHandler);

		try {
			// Admin admin = new
			// AdminFactory().addGroup("stock-demo-analytics-barak").createAdmin();
			final Admin admin = new AdminFactory()
					.addLocator("10.2.241.2:4164").addGroup(
							"stock-demo-analytics").createAdmin();
			admin.getLookupServices().waitFor(1);
			logger.info("[deploy] Found LUS");
			admin.getGridServiceAgents().waitFor(1, 20, TimeUnit.SECONDS);
			final GridServiceAgent[] agents = admin.getGridServiceAgents()
					.getAgents();
			logger.info("Found " + agents.length + " agents");
			if (agents.length < 1) {
				logger.log(Level.SEVERE, "No agent found.");
				System.exit(1);
			}

			/*
			 * Set<Machine> machines = new HashSet<Machine>();
			 * machines.add(agents[0].getMachine());
			 * machines.add(agents[1].getMachine());
			 * machines.add(agents[2].getMachine()); if (machines.size() != 3) {
			 * logger.log(Level.SEVERE,
			 * "the stock demo analytics demo requires 3 machines, each running an agent."
			 * ); System.exit(1); }
			 */

			for (final GridServiceContainer gsc : admin
					.getGridServiceContainers()) {
				gsc.kill();
			}

			for (final ProcessingUnit pu : admin.getProcessingUnits()) {
				pu.undeploy();
			}

			for (final GridServiceManager gsm : admin.getGridServiceManagers()) {
				gsm.kill();
			}

			// Start GSM
			agents[0].startGridService(new GridServiceManagerOptions());

			// Start GSCs
			for (int i = 0; i < 3; i++) {
				agents[i].startGridService(new GridServiceContainerOptions()
						.vmInputArgument("-Dcom.gs.zones=Web_Container")
						// perform GC without stopping the world (semi real-time
						// behavior)
						.vmInputArgument("-XX:+UseConcMarkSweepGC")
						// activate concurrent GC before reaching 80% in which
						// the hight threshold triggers stop-the-world GC
						.vmInputArgument(
								"-XX:CMSInitiatingOccupancyFraction=75")
						// evict young generation in parallel
						.vmInputArgument("-XX:+UseParNewGC").vmInputArgument(
								"-verbose:gc"));
			}

			final ElasticServiceManager elasticServiceManager = admin
					.getElasticServiceManagers().waitForAtLeastOne();
			logger.info("[deploy] Found ESM");
			final String esmThresholdPercentage = "50";

			final ProcessingUnit analyticsDataGridPu = elasticServiceManager
					.deploy(new ElasticDataGridDeployment("stockAnalyticsSpace")
							.elasticScaleHandler(
									new ElasticScaleHandlerConfig(
											"com.gigaspaces.ucs.scaleHandler.UCSElasticScaleHandler")

											// .addProperty("Uri",
											// "http://192.168.10.86/nuova")
											// .addProperty("Username",
											// "config")
											// .addProperty("Password",
											// "config")
											// .addProperty("BladePoolName",
											// "GSPool")

											.addProperty("Uri",
													"https://10.2.140.200/nuova")
											.addProperty("Username",
													"gigaspaces").addProperty(
													"Password", "1Q2w3e4r")
											.addProperty("BladePoolName",
													"GS_SERVERS_POOL")
											.addProperty("lookupGroup",
													"stock-demo-analytics")

											.addProperty("Template",
													"GSTemplates").addProperty(
													"Organization",
													"org-root/org-GigaSpaces")
											.addProperty("serverNamePrefix",
													"GSNode")

											.addProperty("Mode", "2")

											.addProperty("SSHUsername",
													"barakme").addProperty(
													"SSHPassword", "Tojo797")

							)

							// (need to add throttler to clients)
							// the check threshold equals the ESM threshold
							.addContextProperty(
									"space-config.engine.memory_usage.write_only_block_percentage",
									"80")
							.addContextProperty(
									"space-config.engine.memory_usage.write_only_check_percentage",
									esmThresholdPercentage)
							// this is property is not used in ALL-IN-CACHE but
							// it must be lower than the prev threshold.
							.addContextProperty(
									"space-config.engine.memory_usage.low_watermark_percentage",
									"49")

							// reduce yield-time from 1000 to 300 cuts
							// (1000-300)ms*7=4.9 seconds from the active
							// election during fail over (=ESM scale out)
							// we trade active election safety margin for faster
							// performance assuming network latency is less than
							// 300ms.
							.addContextProperty(
									"cluster-config.groups.group.fail-over-policy.active-election.yield-time",
									"300")

							// Only 20k objects stored in memory. Should be
							// enough for this demo.
							// however, when a clear operation is performed it
							// would not have the side effect of tipping the GSC
							// into
							// out of memory condition.
							.addContextProperty(
									"cluster-config.groups.group.repl-policy.redo-log-memory-capacity",
									"20000")

							// this demo never renew object leases so we allow
							// ourselves to disable the default 5 minutes backup
							// expiry delay,
							// in order to let the backup partition clean up
							// expired objects and reduce the memory overhead of
							// expired objects.
							.addContextProperty(
									"space-config.lease_manager.backup_leases_expiration_delay",
									"10000")

							.publicDeploymentIsolation()
							.highlyAvailable(false)

							// perform GC without stopping the world
							.vmInputArgument("-XX:+UseConcMarkSweepGC")
							// activate concurrent GC before reaching 80% in
							// which the hight threshold triggers stop-the-world
							// GC
							// also more likley to invoke garbage collection
							// after a PU has been relocated to another JVM.
							// which
							// would give the ESM a more accurate memory
							// indication.
							.vmInputArgument(
									"-XX:CMSInitiatingOccupancyFraction=40")
							// evict young generation in parallel
							.vmInputArgument("-XX:+UseParNewGC")
							.vmInputArgument("-verbose:gc").capacity("1300m",
									"5200m").maximumJavaHeapSize("1300m")
							.initialJavaHeapSize("1300m")

							.addSla(new MemorySla("60%", 6)));

			logger.info("[deploy] Started deployment of analytics ESM space");
			final GridServiceManager gsm = admin.getGridServiceManagers()
					.waitForAtLeastOne();
			logger.info("[deploy] Found GSM");
			Deploy.waitForSpace(analyticsDataGridPu);
			Deploy.waitForPu(analyticsDataGridPu);
			logger.info("[deploy] Completed deployment of analytics ESM space");

			final ProcessingUnit processorPu = gsm
					.deploy(new ProcessingUnitDeployment(
							"../target/processorPU.jar")
							.addZone("Web_Container")
							.clusterSchema("partitioned-sync2backup")
							.numberOfInstances(3)
							.numberOfBackups(0)
							.maxInstancesPerVM(1)
							.maxInstancesPerMachine(0)
							.setContextProperty(
									"stock.processor.pollingMaxEntries",
									"50000")
							.setContextProperty(
									"stock.processor.pollingTimeoutMilliseconds",
									"500")
							// reduce yield-time from 1000 to 300 cuts
							// (1000-300)ms*7=4.9 seconds from the active
							// election during fail over (=ESM scale out)
							// we trade active election safety margin for faster
							// performance assuming network latency is less than
							// 300ms.
							.setContextProperty(
									"cluster-config.groups.group.fail-over-policy.active-election.yield-time",
									"300"));
			logger.info("[deploy] Started deployment of processorPU");

			final ProcessingUnit analyticsExecutorPu = gsm
					.deploy(new ProcessingUnitDeployment(
							"../target/analyticsPU.jar")
							.setContextProperty(
									"stock.analytics.leaseMilliseconds",
									"360000")
							.setContextProperty("stock.analytics.dummyLength",
									"0")
							.setContextProperty(
									"stock.analytics.samplingPeriodMilliseconds",
									"5000").setContextProperty(
									"stock.analytics.timeWindowMilliseconds",
									"30000").numberOfInstances(1)
							.numberOfBackups(0).addZone("Web_Container"));
			logger.info("[deploy] Started deployment of analyticsPU");
			Deploy.waitForSpace(processorPu);
			Deploy.waitForPrimaryPu(processorPu);
			logger.info("[deploy] Completed deployment of processorPU");
			final ProcessingUnit webPu = gsm
					.deploy(new ProcessingUnitDeployment(
							"../target/StockDemo.war").addZone("Web_Container")
							.maxInstancesPerVM(1).maxInstancesPerMachine(1)
							.numberOfInstances(1));

			logger.info("[deploy] Started deployment of StockDemo");
			final ProcessingUnit processorContainerSlaMonitorPu = gsm
					.deploy(new ProcessingUnitDeployment(
							"../target/ProcessorContainerSlaMonitor.jar")
							.addZone("Web_Container").maxInstancesPerVM(1)
							.maxInstancesPerMachine(1).numberOfInstances(1)
							.setContextProperty("processingunit.name",
									"processorPU"));
			logger
					.info("[deploy] Started deployment of ProcessorContainerSlaMonitor");
			final ProcessingUnit webContainerSlaMonitorPu = gsm
					.deploy(new ProcessingUnitDeployment(
							"../target/WebContainerSlaMonitor.jar").addZone(
							"Web_Container").maxInstancesPerVM(1)
							.maxInstancesPerMachine(1).numberOfInstances(1)
							.setContextProperty("processingunit.name",
									"StockDemo").setContextProperty(
									"machine.name", "Web_Container")
							.setContextProperty("sla.highthreshold", "6")
							.setContextProperty("sla.lowthreshold", "1")
							.setContextProperty("processingunit.maxinstances",
									"4").setContextProperty(
									"processingunit.mininstances", "1"));
			logger
					.info("[deploy] Started deployment of WebContainerSlaMonitor");
			Deploy.waitForPrimaryPu(analyticsExecutorPu);
			logger.info("[deploy] Completed deployment of analyticsPU");
			Deploy.waitForPrimaryPu(webPu);
			logger.info("[deploy] Completed deployment of StockDemo");
			Deploy.waitForPrimaryPu(processorContainerSlaMonitorPu);
			logger
					.info("[deploy] Completed deployment of ProcessorContainerSlaMonitor");
			Deploy.waitForPrimaryPu(webContainerSlaMonitorPu);
			logger
					.info("[deploy] Completed deployment of WebContainerSlaMonitor");

			// this starts the actual demo
			final ProcessingUnit feederPu = gsm
					.deploy(new ProcessingUnitDeployment(
							"../target/feederPU.jar").addZone("Web_Container")
							.maxInstancesPerVM(1).maxInstancesPerMachine(1)
							.numberOfInstances(1).numberOfBackups(1)
							.setContextProperty("stock.feeder.batchsize",
									"1000").setContextProperty(
									"stock.feeder.intervalmilliseconds", "500")
							.setContextProperty(
									"stock.feeder.numberOfWritingThreads", "1")
							.setContextProperty(
									"stock.feeder.perMinuteStockVolatility",
									"0.01").setContextProperty(
									"stock.feeder.filename", "stocks.txt")
							// .setContextProperty("stock.feeder.filename","stocks-small.txt")

							// reduce yield-time from 1000 to 300 cuts
							// (1000-300)ms*7=4.9 seconds from the active
							// election during fail over (=ESM scale out)
							// we trade active election safety margin for faster
							// performance assuming network latency is less than
							// 300ms.
							.setContextProperty(
									"cluster-config.groups.group.fail-over-policy.active-election.yield-time",
									"300"));
			logger.info("[deploy] Started deployment of feederPU");
			Deploy.waitForPrimaryPu(feederPu);
			logger.info("[deploy] Completed deployment of feederPU");
		} catch (final Exception e) {
			logger.log(Level.SEVERE, "Error during PU deployment", e);
			System.exit(1);
		}
		System.exit(0);
	}

	private static void waitForPrimaryPu(final ProcessingUnit pu) {
		pu.waitFor(pu.getNumberOfInstances());
		logger.info("Finished deploying primaries of pu " + pu.getName());
	}

	private static void waitForPu(final ProcessingUnit pu) {
		pu.waitFor(pu.getTotalNumberOfInstances());
		logger.info("Finished deploying pu " + pu.getName());
	}

	private static void waitForSpace(final ProcessingUnit pu) {
		final Space space = pu.waitForSpace();
		logger.info("Finished deploying space " + space.getName());
	}
}
