package com.gigaspaces.ucs.deploy;

import java.io.File;
import java.io.IOException;
import java.net.ServerSocket;

import org.eclipse.jetty.server.Handler;
import org.eclipse.jetty.server.Server;
import org.eclipse.jetty.server.handler.DefaultHandler;
import org.eclipse.jetty.server.handler.HandlerList;
import org.eclipse.jetty.server.handler.ResourceHandler;
import org.eclipse.jetty.server.nio.SelectChannelConnector;

import com.gigaspaces.ucs.scaleHandler.UCSElasticScaleHandler;

/*******
 * An embedded HTTP server that serves file from a given directory. Uses the
 * Jetty HTTP server.
 * 
 * Not used, but left in the code for possible future uses.
 * 
 * @author barakme
 * 
 */
public class HttpServer {

	private static java.util.logging.Logger logger = java.util.logging.Logger.getLogger(UCSElasticScaleHandler.class
			.getName());	

	private final String docBase;

	private int port;

	private Server server;

	/********
	 * Constructor for the http server. Call start() for the server to start
	 * handling requests.
	 * 
	 * @param docBase
	 *            base directory from which files are served.
	 * @param port
	 *            port for the server. If 0, a random free port is chosen.
	 */
	public HttpServer(final String docBase, final int port) {
		this.docBase = docBase;
		this.port = port;
	}

	private int getFreePort() throws IOException {
		ServerSocket sock = null;
		try {
			sock = new ServerSocket(0);
			return sock.getLocalPort();
		} finally {

			try {
				sock.close();
			} catch (final IOException e) {
				// ignore
			}

		}

	}

	public int getPort() {
		return port;
	}

	private void initDocBase(final String docBase) {
		final File file = new File(docBase);
		if (file.exists()) {
			if (file.isDirectory()) {
				return;
			}

			throw new IllegalArgumentException("docBase for ESM HTTP server is set to " + docBase
					+ ", which is a file. It should be a directory!");
		}

		if (!file.mkdirs()) {
			throw new IllegalArgumentException("Failed to create ESM HTTP Server docBase at: " + docBase);
		}
	}

	/*****
	 * Stops this HTTP server.
	 */
	public void shutdown() {
		this.server.destroy();

	}

	/******
	 * Start handling HTTP requests.
	 * 
	 * @throws IOException
	 *             if server could not start.
	 */
	public void start() throws IOException {
		if (this.port == 0) {
			this.port = getFreePort();
			logger.info("Chose port " + port + " for Http Server");
		}

		// make sure the base directory exists, or create it
		initDocBase(docBase);

		server = new Server();
		final SelectChannelConnector connector = new SelectChannelConnector();
		connector.setPort(port);
		server.addConnector(connector);

		final ResourceHandler resourceHandler = new ResourceHandler();
		resourceHandler.setDirectoriesListed(true);
		// resource_handler.setWelcomeFiles(new String[] { "index.html" });

		resourceHandler.setResourceBase(docBase);

		final HandlerList handlers = new HandlerList();
		handlers.setHandlers(new Handler[] { resourceHandler, new DefaultHandler() });
		server.setHandler(handlers);

		try {
			server.start();
			logger.info("Started HTTP Server");
		} catch (final Exception e) {
			throw new IOException("Failed to start HTTP server", e);
		}
	}
}
