package com.gigaspaces.ucs.service;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.Serializable;
import java.io.StringReader;
import java.io.StringWriter;
import java.io.UnsupportedEncodingException;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.net.URI;
import java.net.URISyntaxException;
import java.security.cert.CertificateException;
import java.security.cert.X509Certificate;
import java.util.ArrayList;
import java.util.Date;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;
import java.util.concurrent.Executors;
import java.util.concurrent.ScheduledExecutorService;
import java.util.concurrent.ScheduledFuture;
import java.util.concurrent.TimeUnit;

import javax.net.ssl.SSLContext;
import javax.net.ssl.TrustManager;
import javax.net.ssl.X509TrustManager;
import javax.xml.bind.JAXBContext;
import javax.xml.bind.JAXBElement;
import javax.xml.bind.JAXBException;
import javax.xml.bind.Marshaller;
import javax.xml.bind.Unmarshaller;
import javax.xml.namespace.QName;
import javax.xml.transform.stream.StreamSource;

import org.apache.http.HttpEntity;
import org.apache.http.HttpResponse;
import org.apache.http.client.HttpClient;
import org.apache.http.client.methods.HttpPost;
import org.apache.http.conn.ClientConnectionManager;
import org.apache.http.conn.scheme.Scheme;
import org.apache.http.conn.scheme.SchemeRegistry;
import org.apache.http.conn.ssl.SSLSocketFactory;
import org.apache.http.entity.StringEntity;
import org.apache.http.impl.client.DefaultHttpClient;

import com.gigaspaces.internal.backport.java.util.Arrays;
import com.gigaspaces.ucs.domain.in.AaaKeepAlive;
import com.gigaspaces.ucs.domain.in.AaaLogin;
import com.gigaspaces.ucs.domain.in.AaaRefresh;
import com.gigaspaces.ucs.domain.in.ConfigConfMo;
import com.gigaspaces.ucs.domain.in.ConfigConfMos;
import com.gigaspaces.ucs.domain.in.ConfigMap;
import com.gigaspaces.ucs.domain.in.ConfigResolveChildren;
import com.gigaspaces.ucs.domain.in.ConfigResolveClass;
import com.gigaspaces.ucs.domain.in.LsBinding;
import com.gigaspaces.ucs.domain.in.LsClone;
import com.gigaspaces.ucs.domain.in.LsInstantiateTemplate;
import com.gigaspaces.ucs.domain.in.LsPower;
import com.gigaspaces.ucs.domain.in.LsRequirement;
import com.gigaspaces.ucs.domain.in.LsServer;
import com.gigaspaces.ucs.domain.in.LsbootDef;
import com.gigaspaces.ucs.domain.in.LsbootLocalStorage;
import com.gigaspaces.ucs.domain.in.LsbootStorage;
import com.gigaspaces.ucs.domain.in.ConfigMap.Pair;
import com.gigaspaces.ucs.domain.out.ComputeBlade;
import com.gigaspaces.ucs.domain.out.ComputePool;
import com.gigaspaces.ucs.domain.out.ComputePooledSlot;
import com.gigaspaces.ucs.domain.out.ConfigConfig;
import com.gigaspaces.ucs.domain.out.ConfigSet;
import com.gigaspaces.ucs.domain.out.MacpoolPooled;

/**********
 * This is the main service class for the UCS Manager client API. The service
 * class should only be instantiated once as construction is heavy due to the
 * JAXBContext initialization. The service class is thread-safe, so it can be
 * cached and re-used.
 * 
 * Note that if you plan to use the service class for a while, the keep alive
 * method should be used to make sure that the login session does not expire.
 * 
 * @author barakme
 * 
 */
public class Service {

	// UCS Element Status constants
	private static final String STATUS_MODIFIED = "modified";
	private static final String STATUS_CREATED_MODIFIED = "created,modified";
	private static final String STATUS_DELETED = "deleted";
	public static final String STATUS_CREATED = "created";

	// UCS Class IDs
	public static final String CLASS_COMPUTE_BLADE = "computeBlade";
	public static final String CLASS_LS_SERVER = "lsServer";

	// Logger
	private static java.util.logging.Logger logger = java.util.logging.Logger.getLogger(Service.class.getName());

	private static final int HTTPS_PORT = 443;

	/***********
	 * Main runner methods for quick tests.
	 * 
	 * @param args
	 *            the command line arguments.
	 * @throws UCSServiceException
	 *             In case of an error.
	 * @throws URISyntaxException
	 *             In case the server URL is malformed.
	 */
	public static void main(final String[] args) throws UCSServiceException, URISyntaxException {

		final Service service = new Service(new URI("http://10.122.170.131/nuova"));
		service.login("admin", "P@55w0rd!");
		final List<MacpoolPooled> list = service.getMacPool("default");
		System.out.println(list.size());

		// service.login("config", "config");
		// final Service service = new Service(new
		// URI("https://127.0.10.1/nuova"));
		// service.login("gigaspaces", "1Q2w3e4r");

		// service.launchSessionKeepAliveTask();
		// service.disassociateServer("org-root/ls-GSServerProfile");

		// service.shutdownServer("org-root/org-GigaSpaces/ls-GSNode2", true);
		/*
		 * final List<com.gigaspaces.ucs.domain.out.LsServer> allprofiles =
		 * service .getAllServiceProfilesInOrg("root/GigaSpaces"); for (final
		 * com.gigaspaces.ucs.domain.out.LsServer profile : allprofiles) { final
		 * String assoc = profile.getAssocState(); final boolean power =
		 * service.isProfilePoweredOn(profile.getDn()); if(power) {
		 * System.out.println("shutting down " + profile.getDn());
		 * service.shutdownServer(profile.getDn(), false); } else {
		 * System.out.println("starting " + profile.getDn());
		 * service.bootServer(profile.getDn()); } }
		 */

		/*
		 * List<com.gigaspaces.ucs.domain.out.LsServer> profiles =
		 * service.getAllUnassociatedServiceProfilesInOrg("root/GigaSpaces");
		 * 
		 * 
		 * com.gigaspaces.ucs.domain.out.LsServer server =
		 * service.associateProfileToBladePool(profiles.get(0).getDn(),
		 * "GSPool");
		 */

	}

	// JAXB items
	private Unmarshaller unmarshaller;

	private Marshaller marshaller;
	private Unmarshaller eventUnmarshaller;
	// HTTP Items
	private final URI uri;
	private final HttpClient httpClient;
	// Session items
	private String cookie;

	private Long refreshPeriod;
	private Date lastAccessed;

	private final ScheduledExecutorService keepAliveExecutor = Executors.newScheduledThreadPool(1);

	private ScheduledFuture<?> keepAliveTask = null;

	/*********
	 * Constructor for the service facade.
	 * 
	 * @param host
	 *            Host name for the UCS manager.
	 * @param port
	 *            port number of the UCS manager.
	 * @param url
	 *            URL of the UCS manager.
	 * @throws URISyntaxException
	 *             If the URL is illegal.
	 * @throws UCSServiceException
	 *             in case of an error.
	 */
	public Service(final String host, final int port, final String url) throws URISyntaxException, UCSServiceException {
		this(host, port, url, true);
	}

	/*********
	 * Constructor for the service facade.
	 * 
	 * @param host
	 *            Host name for the UCS manager.
	 * @param port
	 *            port number of the UCS manager.
	 * @param url
	 *            URL of the UCS manager.
	 * @param fastBoot
	 *            true if the JAXB fast boot option should be used.
	 * @throws URISyntaxException
	 *             If the URL is illegal.
	 * @throws UCSServiceException
	 *             in case of an error.
	 */
	public Service(final String host, final int port, final String url, final boolean fastBoot)
			throws URISyntaxException, UCSServiceException {

		this(new URI((port == HTTPS_PORT ? "https" : "http"), null, host, port, url, null, null), fastBoot);
	}

	/*********
	 * Constructor for the UCS Manager service facade.
	 * 
	 * @param uri
	 *            Full URI for the UCS Manager, including protocol, host and
	 *            port.
	 * @throws URISyntaxException .
	 * @throws UCSServiceException .
	 */
	public Service(final URI uri) throws UCSServiceException, URISyntaxException {
		this(uri, true);
	}

	/*********
	 * Constructor for the UCS Manager service facade.
	 * 
	 * @param uri
	 *            Full URI for the UCS Manager, including protocol, host and
	 *            port.
	 * @param fastBoot
	 *            true if the fast boot JAXB option should be used.
	 * @throws URISyntaxException .
	 * @throws UCSServiceException .
	 */
	public Service(final URI uri, final boolean fastBoot) throws UCSServiceException, URISyntaxException {
		this.uri = uri;

		setFastBoot(fastBoot);

		logger.info("Starting UCS Service. " + "Initialization may take a few seconds.");
		final long begin = System.currentTimeMillis();

		try {
			final JAXBContext jaxbContextIn = JAXBContext.newInstance("com.gigaspaces.ucs.domain.in");

			/*
			 * JAXBContext jaxbContextIn =
			 * JAXBContext.newInstance(AaaLogin.class,
			 * LsInstantiateTemplate.class, ConfigResolveChildren.class);
			 */

			marshaller = jaxbContextIn.createMarshaller();
			eventUnmarshaller = jaxbContextIn.createUnmarshaller();

			final JAXBContext jaxbContextOut = JAXBContext.newInstance("com.gigaspaces.ucs.domain.out");
			/*
			 * JAXBContext jaxbContextOut = JAXBContext.newInstance(
			 * com.gigaspaces.ucs.domain.out.AaaLogin.class,
			 * com.gigaspaces.ucs.domain.out.LsInstantiateTemplate.class,
			 * com.gigaspaces.ucs.domain.out.ConfigResolveChildren.class);
			 */
			unmarshaller = jaxbContextOut.createUnmarshaller();
		} catch (final JAXBException e) {
			throw new UCSServiceException("Failed to prepare XML marhaller and unmarshaller: " + e, e);
		}
		final long end = System.currentTimeMillis();
		logger.info("Service startup time: " + (end - begin) + " ms");

		httpClient = useTrustingTrustManager(new DefaultHttpClient());

	}

	private com.gigaspaces.ucs.domain.out.LsPower applyPowerState(final String serverDn, final String state)
			throws UCSServiceException {

		final String dn = serverDn + "/power";

		final LsPower power = new LsPower();
		power.setDn(dn);
		power.setState(state);
		final com.gigaspaces.ucs.domain.out.LsPower res = configConfMo(power, dn,
				com.gigaspaces.ucs.domain.out.LsPower.class);

		if (!res.getState().equals(state)) {
			throw new UCSServiceException(state + " state was not applied to server!");
		}

		return res;

	}

	/******
	 * Associate profile to specific blade.
	 * 
	 * @param profileDn
	 *            the service profile DN.
	 * @param bladeDn
	 *            the blade DN.
	 * @return the modified profile.
	 * @throws UCSServiceException .
	 */
	public com.gigaspaces.ucs.domain.out.LsServer associateProfileToBlade(final String profileDn, final String bladeDn)
			throws UCSServiceException {
		final LsBinding binding = new LsBinding();
		binding.setPnDn(bladeDn);
		binding.setStatus(STATUS_CREATED_MODIFIED);

		final LsServer profile = new LsServer();
		profile.getContent().add(wrapObjectInJAXBElement(binding));
		profile.setDn(profileDn);
		profile.setStatus(STATUS_CREATED_MODIFIED);

		final com.gigaspaces.ucs.domain.out.LsServer ret = configConfMos(profile, profileDn,
				com.gigaspaces.ucs.domain.out.LsServer.class);

		if (!ret.getStatus().equalsIgnoreCase(STATUS_MODIFIED)) {
			throw new UCSServiceException("Expected to recieve status modified, but got: " + ret.getStatus());
		}
		return ret;

	}

	/******
	 * Associates a service profile to a physical blade from a pool of blades.
	 * 
	 * @param profileDn
	 *            the service profile DN.
	 * @param poolName
	 *            the blade pool name.
	 * @return the modified service profile.
	 * @throws UCSServiceException .
	 */
	public com.gigaspaces.ucs.domain.out.LsServer associateProfileToBladePool(final String profileDn,
			final String poolName) throws UCSServiceException {

		final LsRequirement req = new LsRequirement();
		req.setName(poolName);
		req.setRn("pn-req");
		req.setQualifier("");

		final LsServer server = new LsServer();
		server.getContent().add(wrapObjectInJAXBElement(req));
		server.setDn(profileDn);

		final com.gigaspaces.ucs.domain.out.LsServer res = configConfMos(server, profileDn,
				com.gigaspaces.ucs.domain.out.LsServer.class);

		if (res == null) {
			throw new UCSServiceException("Expected to get a server in the response, but received null");
		}
		if (!res.getStatus().contains("modified")) {
			throw new UCSServiceException("Expected to recieve a modified server status, but got: " + res.getStatus());
		}

		return res;

	}

	/********
	 * Powersw up a server.
	 * 
	 * @param dn
	 *            the service profile DN.
	 * @return the modified power state.
	 * @throws UCSServiceException .
	 */
	public com.gigaspaces.ucs.domain.out.LsPower bootServer(final String dn) throws UCSServiceException {
		return applyPowerState(dn, "up");
	}

	private void checkForErrorResponse(final Object obj) throws UCSServiceException {
		Long errorCode = null;
		String errorDescr = null;
		String invocationResult = null;

		try {
			final Method getterCode = obj.getClass().getMethod("getErrorCode", (Class<?>[]) null);
			errorCode = (Long) getterCode.invoke(obj, (Object[]) null);
			if (errorCode == null) {
				return; // no error
			}

			final Method getterDesc = obj.getClass().getMethod("getErrorDescr", (Class<?>[]) null);
			errorDescr = (String) getterDesc.invoke(obj, (Object[]) null);

			final Method getterResult = obj.getClass().getMethod("getInvocationResult", (Class<?>[]) null);
			invocationResult = (String) getterResult.invoke(obj, (Object[]) null);

		} catch (final SecurityException e) {
			// ignore - the values will be null and an exception will be thrown
			// after the catch block.
		} catch (final NoSuchMethodException e) {
			// ignore - the values will be null and an exception will be thrown
			// after the catch block.
		} catch (final IllegalAccessException e) {
			// ignore - the values will be null and an exception will be thrown
			// after the catch block.
		} catch (final IllegalArgumentException e) {
			// ignore - the values will be null and an exception will be thrown
			// after the catch block.
		} catch (final InvocationTargetException e) {
			// ignore - the values will be null and an exception will be thrown
			// after the catch block.
		}
		throw new UCSServiceException("Recieved an error response from UCS Manager: Code - " + errorCode
				+ ", Result - " + invocationResult + ", Description - " + errorDescr);

	}

	/********
	 * Clones a service profile.
	 * 
	 * @param origDN
	 *            the DN of the original service profile.
	 * @param org
	 *            The org name.
	 * @param newProfileName
	 *            the new profile name.
	 * @return the logical server element for the cloned profile.
	 * @throws UCSServiceException
	 *             if there was a problem while cloning the service profile.
	 */
	public com.gigaspaces.ucs.domain.out.LsServer cloneServiceProfile(final String origDN, final String org,
			final String newProfileName) throws UCSServiceException {
		final LsClone clone = new LsClone();
		clone.setCookie(cookie);
		clone.setDn(origDN);
		clone.setInHierarchical(Boolean.toString(false));
		clone.setInTargetOrg(org);
		clone.setInServerName(newProfileName);

		final com.gigaspaces.ucs.domain.out.LsClone res = submit(clone, com.gigaspaces.ucs.domain.out.LsClone.class);

		final ConfigConfig config = getElementFromJAXBList(res.getContent(), ConfigConfig.class);
		return (com.gigaspaces.ucs.domain.out.LsServer) config.getManagedObject().getValue();

	}

	/*********
	 * Calls the Service Manager configConfMo method.
	 * 
	 * @param <T>
	 *            the expected return type.
	 * @param confObject
	 *            the configuration object to send.
	 * @param dn
	 *            the DN of the element where the configuration will be applied.
	 * @param retType
	 *            the expeted return type.
	 * @return the method response.
	 * @throws UCSServiceException .
	 */
	@SuppressWarnings("unchecked")
	public <T> T configConfMo(final Object confObject, final String dn, final Class<T> retType)
			throws UCSServiceException {

		final com.gigaspaces.ucs.domain.in.ConfigConfig cc = new com.gigaspaces.ucs.domain.in.ConfigConfig();
		cc.setManagedObject(wrapObjectInJAXBElement(confObject));

		final ConfigConfMo conf = new ConfigConfMo();
		conf.setCookie(this.cookie);
		conf.setInHierarchical(Boolean.toString(false));
		conf.setDn(dn);
		conf.getContent().add(
				new JAXBElement<com.gigaspaces.ucs.domain.in.ConfigConfig>(new QName("inConfig"),
						com.gigaspaces.ucs.domain.in.ConfigConfig.class, cc));

		final com.gigaspaces.ucs.domain.out.ConfigConfMo resp = submit(conf,
				com.gigaspaces.ucs.domain.out.ConfigConfMo.class);

		final com.gigaspaces.ucs.domain.out.ConfigConfig respCC = getElementFromJAXBList(resp.getContent(),
				com.gigaspaces.ucs.domain.out.ConfigConfig.class);

		if (respCC.getManagedObject() == null) {
			throw new UCSServiceException("No configuration object found in configConfMo response");
		}

		final T resultEntiry = (T) respCC.getManagedObject().getValue();

		return resultEntiry;

	}

	/*********
	 * The configConfMos method of the UCS Manager API. Most configuration
	 * methods delegate to this method. If the service API does not have a
	 * method matching your needs, you can always construct your request using
	 * the domain objects and use this method to send the request to the UCS
	 * manager.
	 * 
	 * @param <T>
	 *            The top level return type element sent back by the server
	 *            (from the domain.out package).
	 * @param confObject
	 *            the request object (from the domain.in package).
	 * @param dn
	 *            the DN of the request.
	 * @param retType
	 *            The class of the return type.
	 * @return the return value.
	 * @throws UCSServiceException
	 *             if there was a problem while executing the request.
	 */
	@SuppressWarnings("unchecked")
	public <T> T configConfMos(final Object confObject, final String dn, final Class<T> retType)
			throws UCSServiceException {

		// create the pair
		final Pair pair = new Pair();
		pair.setKey(dn);
		pair.setManagedObject(wrapObjectInJAXBElement(confObject));

		final ConfigMap cm = new ConfigMap();
		cm.getPair().add(pair);

		final ConfigConfMos conf = new ConfigConfMos();
		conf.setCookie(this.cookie);
		conf.setInHierarchical(Boolean.toString(false));
		// JAXB behaves oddly here. The content field ConfigConfMos should be
		// names
		// inConfigs automatically, and the annotation lists it, but it doesn't
		// happen.
		// So we have to set it here explicitly. This is probably due to the xsd
		// using
		// xsd:any tags liberally. JAXB does not handle that well.
		conf.getContent().add(new JAXBElement<ConfigMap>(new QName("inConfigs"), ConfigMap.class, cm));
		// conf.getContent().add(wrapObjectInJAXBElement(" "));

		final com.gigaspaces.ucs.domain.out.ConfigConfMos resp = submit(conf,
				com.gigaspaces.ucs.domain.out.ConfigConfMos.class);

		final com.gigaspaces.ucs.domain.out.ConfigMap respCm = getElementFromJAXBList(resp.getContent(),
				com.gigaspaces.ucs.domain.out.ConfigMap.class);

		if (respCm.getPair().size() == 0) {
			throw new UCSServiceException("No configuration pair found in configConfMos response");
		}

		final T resultEntiry = (T) respCm.getPair().get(0).getManagedObject().getValue();

		return resultEntiry;

	}

	/*********
	 * The configResolveChildren method of the UCS manager.
	 * 
	 * @param <T>
	 *            the expected return type.
	 * @param classId
	 *            the class ID being queried.
	 * @param inDn
	 *            the DN being queried.
	 * @param inHierarchicha
	 *            if the request should be applied to the hierarchy.
	 * @return the method response.
	 * @throws UCSServiceException .
	 */
	@SuppressWarnings("unchecked")
	public <T> List<T> configResolveChildren(final String classId, final String inDn, final boolean inHierarchicha)
			throws UCSServiceException {
		logger.info("COnfigResolveChildren: classId: " + classId + ", DN: " + inDn);
		final ConfigResolveChildren query = new ConfigResolveChildren();
		query.setCookie(this.cookie);
		query.setClassId(classId);
		query.setInDn(inDn);
		query.setInHierarchical(Boolean.toString(inHierarchicha));

		final com.gigaspaces.ucs.domain.out.ConfigResolveChildren queryRes = submit(query,
				com.gigaspaces.ucs.domain.out.ConfigResolveChildren.class);

		for (final Serializable obj : queryRes.getContent()) {

			if (obj instanceof JAXBElement<?>) {
				final JAXBElement<ConfigSet> set = (JAXBElement<ConfigSet>) obj;
				final List<JAXBElement<?>> list = set.getValue().getManagedObject();
				final List<T> domainList = getDomainObjectFromJaxList(list);
				return domainList;
			}

		}

		return null;

	}

	/************
	 * The configResolveClass method of the UCS Manager API.
	 * 
	 * @param <T>
	 *            expected return type.
	 * @param classId
	 *            the class ID.
	 * @param inHierarchichal
	 *            .
	 * @return .
	 * @throws UCSServiceException .
	 */
	@SuppressWarnings("unchecked")
	public <T> List<T> configResolveClass(final String classId, final boolean inHierarchichal)
			throws UCSServiceException {
		final ConfigResolveClass query = new ConfigResolveClass();
		query.setCookie(this.cookie);
		query.setClassId(classId);
		query.setInHierarchical(Boolean.toString(inHierarchichal));

		final com.gigaspaces.ucs.domain.out.ConfigResolveChildren queryRes = submit(query,
				com.gigaspaces.ucs.domain.out.ConfigResolveChildren.class);

		for (final Serializable obj : queryRes.getContent()) {

			if (obj instanceof JAXBElement<?>) {
				final JAXBElement<ConfigSet> set = (JAXBElement<ConfigSet>) obj;
				final List<JAXBElement<?>> list = set.getValue().getManagedObject();
				final List<T> domainList = getDomainObjectFromJaxList(list);
				return domainList;
			}

		}

		return null;

	}

	/*******
	 * A useful debugging method.
	 * 
	 * @param name
	 *            the name of the profile.
	 * @return the created profile.
	 * @throws UCSServiceException .
	 */
	public com.gigaspaces.ucs.domain.out.LsServer createTestServiceProfile(final String name)
			throws UCSServiceException {
		final LsbootLocalStorage localStorage = new LsbootLocalStorage();
		localStorage.setRn("local-storage");
		localStorage.setStatus(STATUS_CREATED);

		final LsbootStorage bootStorage = new LsbootStorage();
		bootStorage.getContent().add(wrapObjectInJAXBElement(localStorage));
		bootStorage.setOrder("1");
		bootStorage.setRn("storage");

		final LsbootDef boot = new LsbootDef();
		boot.getContent().add(wrapObjectInJAXBElement(bootStorage));
		boot.setEnforceVnicName("no");
		boot.setRebootOnUpdate("no");
		boot.setRn("boot-policy");

		final LsServer server = new LsServer();
		server.getContent().add(wrapObjectInJAXBElement(boot));
		server.setDn("org-root/ls-" + name);
		server.setName(name);
		server.setStatsPolicyName("default");
		server.setStatus(STATUS_CREATED);
		server.setUuid("0");

		final Pair pair = new Pair();
		pair.setManagedObject(wrapObjectInJAXBElement(server));
		pair.setKey("org-root/ls-" + name);

		final ConfigMap cm = new ConfigMap();
		cm.getPair().add(pair);

		final ConfigConfMos conf = new ConfigConfMos();
		conf.getContent().add(new JAXBElement<ConfigMap>(new QName("inConfigs"), ConfigMap.class, cm));

		final com.gigaspaces.ucs.domain.out.ConfigConfMos res = submit(conf,
				com.gigaspaces.ucs.domain.out.ConfigConfMos.class);

		final com.gigaspaces.ucs.domain.out.ConfigMap respCm = getElementFromJAXBList(res.getContent(),
				com.gigaspaces.ucs.domain.out.ConfigMap.class);
		final com.gigaspaces.ucs.domain.out.LsServer resultServer = (com.gigaspaces.ucs.domain.out.LsServer) respCm
				.getPair().get(0).getManagedObject().getValue();

		if (!resultServer.getStatus().equals(STATUS_CREATED)) {
			throw new UCSServiceException("Expected status created in response but received: "
					+ resultServer.getStatus());
		}

		return resultServer;

	}

	/**********
	 * Deletes a service profile.
	 * 
	 * @param dn
	 *            The profile DN.
	 * @return the modified profile.
	 * @throws UCSServiceException .
	 */
	public com.gigaspaces.ucs.domain.out.LsServer deleteServiceProfile(final String dn) throws UCSServiceException {
		final LsServer server = new LsServer();
		server.setDn(dn);
		server.setStatus(STATUS_DELETED);

		final com.gigaspaces.ucs.domain.out.LsServer ret = configConfMos(server, dn,
				com.gigaspaces.ucs.domain.out.LsServer.class);

		if (!ret.getStatus().equals(STATUS_DELETED)) {
			throw new UCSServiceException("Expected to receive status deleted, got: " + ret.getStatus());
		}

		return ret;
	}

	/*****
	 * Disassociate a profile from a physical blade.
	 * 
	 * @param serverDN
	 *            the service profile DN.
	 * @return the modified profile.
	 * @throws UCSServiceException .
	 */
	public com.gigaspaces.ucs.domain.out.LsServer disassociateServer(final String serverDN) throws UCSServiceException {
		// Create the binding
		final LsBinding binding = new LsBinding();
		binding.setRn("pn");
		binding.setStatus(STATUS_DELETED);

		// create logical server and add Binding
		final LsServer server = new LsServer();
		server.setDn(serverDN);
		server.getContent().add(wrapObjectInJAXBElement(binding));

		final com.gigaspaces.ucs.domain.out.LsServer ret = configConfMos(server, serverDN,
				com.gigaspaces.ucs.domain.out.LsServer.class);

		return ret;

	}

	/*****
	 * Returns all blades.
	 * 
	 * @return the blades.
	 * @throws UCSServiceException .
	 */
	public List<ComputeBlade> getAllComputeBlades() throws UCSServiceException {
		final ConfigResolveClass req = new ConfigResolveClass();
		req.setInHierarchical(Boolean.toString(false));
		req.setCookie(this.cookie);
		req.setClassId(CLASS_COMPUTE_BLADE);

		final List<ComputeBlade> blades = configResolveClass(CLASS_COMPUTE_BLADE, false);

		return blades;

	}

	/******
	 * Returns all service profiles.
	 * 
	 * @return the profiles.
	 * @throws UCSServiceException .
	 */
	public List<com.gigaspaces.ucs.domain.out.LsServer> getAllServiceProfiles() throws UCSServiceException {
		final List<com.gigaspaces.ucs.domain.out.LsServer> profiles = configResolveClass(CLASS_LS_SERVER, false);

		return profiles;
	}

	/********
	 * Returns all service profiles that are in an organization.
	 * 
	 * @param org
	 *            the organization DN.
	 * @return the profiles.
	 * @throws UCSServiceException .
	 */
	public List<com.gigaspaces.ucs.domain.out.LsServer> getAllServiceProfilesInOrg(final String org)
			throws UCSServiceException {
		final String normalizedOrg = normalizeOrgName(org);
		final List<com.gigaspaces.ucs.domain.out.LsServer> allprofiles = getAllServiceProfiles();
		final List<com.gigaspaces.ucs.domain.out.LsServer> profiles = 
			new LinkedList<com.gigaspaces.ucs.domain.out.LsServer>();
		for (final com.gigaspaces.ucs.domain.out.LsServer lsServer : allprofiles) {
			if (lsServer.getType().equals("instance") && (lsServer.getDn().indexOf(normalizedOrg) >= 0)) {
				profiles.add(lsServer);
			}
		}

		return profiles;
	}

	/********
	 * Returns all unassociated profiles in an organization.
	 * 
	 * @param org
	 *            the org DN.
	 * @return the profiles.
	 * @throws UCSServiceException .
	 */
	public List<com.gigaspaces.ucs.domain.out.LsServer> getAllUnassociatedServiceProfilesInOrg(final String org)
			throws UCSServiceException {
		final String normalizedOrg = normalizeOrgName(org);
		final List<com.gigaspaces.ucs.domain.out.LsServer> allprofiles = getAllServiceProfiles();
		final List<com.gigaspaces.ucs.domain.out.LsServer> profiles = 
			new LinkedList<com.gigaspaces.ucs.domain.out.LsServer>();
		for (final com.gigaspaces.ucs.domain.out.LsServer lsServer : allprofiles) {
			if (lsServer.getType().equals("instance") && lsServer.getAssocState().equals("unassociated")
					&& lsServer.getDn().startsWith(normalizedOrg)) {
				profiles.add(lsServer);
			}
		}

		return profiles;
	}

	/*****
	 * Gets a blade by its name.
	 * 
	 * @param dn
	 *            the blade name.
	 * @return the blade.
	 * @throws UCSServiceException .
	 */
	public ComputeBlade getComputeBlade(final String dn) throws UCSServiceException {
		final List<ComputeBlade> blades = getAllComputeBlades();
		for (final ComputeBlade blade : blades) {
			if (blade.getDn().equals(dn)) {
				return blade;
			}
		}
		return null;
	}

	/************
	 * Gets a blade pool by its name.
	 * 
	 * @param poolName
	 *            pool name.
	 * @return the pool.
	 * @throws UCSServiceException .
	 */
	public List<ComputePooledSlot> getComputePool(final String poolName) throws UCSServiceException {
		final List<ComputePooledSlot> res = configResolveChildren("computePooledSlot", "org-root/compute-pool-"
				+ poolName, false);
		return res;

	}

	public String getCookie() {
		return cookie;
	}

	@SuppressWarnings("unchecked")
	private <T> List<T> getDomainObjectFromJaxList(final List<JAXBElement<?>> list) {
		final List<T> res = new ArrayList<T>(list.size());
		for (final JAXBElement<?> elem : list) {
			res.add((T) elem.getValue());
		}
		return res;
	}

	@SuppressWarnings("unchecked")
	private <T> T getElementFromJAXBList(final List<Serializable> list, 
			final Class<T> type) throws UCSServiceException {

		for (final Serializable temp : list) {
			if (temp instanceof JAXBElement<?>) {
				final JAXBElement<T> element = (JAXBElement<T>) temp;
				return element.getValue();
			}
		}
		throw new UCSServiceException("Could not find element of type " + type.getSimpleName() + " in response");
	}

	/*****
	 * Returns the time that this service last accessed the server.
	 * 
	 * @return the time that this service last accessed the server.
	 */
	public Date getLastAccessed() {
		return new Date(lastAccessed.getTime());
	}

	/****
	 * Gets a MAC pool.
	 * 
	 * @param macPoolDN
	 *            the pool DN.
	 * @return the pool.
	 * @throws UCSServiceException .
	 */
	public List<MacpoolPooled> getMacPool(final String macPoolDN) throws UCSServiceException {
		String actualPoolName = macPoolDN;
		if (!macPoolDN.startsWith("org-root")) {
			actualPoolName = "org-root/mac-pool-" + macPoolDN;
		}
		final List<MacpoolPooled> macs = configResolveChildren("macpoolPooled", actualPoolName, false);

		return macs;

	}

	/********
	 * Returns the MAC addresses used by this server. List will be empty if no
	 * MAC address in this pool is used by this server.
	 * 
	 * @param macPoolDN
	 *            the MAC pool DN.
	 * @param serverDN
	 *            the profile DN.
	 * @return List of MAC addresses.
	 * @throws UCSServiceException .
	 */
	public List<String> getMacsForServer(final String macPoolDN, final String serverDN) throws UCSServiceException {
		logger.info("Getting MAC Addresses for server: " + serverDN + " from MAC pool: " + macPoolDN);
		final List<MacpoolPooled> pool = getMacPool(macPoolDN);
		logger.info("Found " + pool.size() + " entries in pool. Trying to match to server");
		final List<String> res = new ArrayList<String>();
		for (final MacpoolPooled macpoolPooled : pool) {
			final String dn = macpoolPooled.getAssignedToDn();
			if ((dn != null) && (dn.length() > 0)) {
				if (dn.contains(serverDN)) {
					res.add(macpoolPooled.getId());
				}
			}
		}

		return res;

	}

	/********
	 * Returns all compute pool names from the UCS Manager.
	 * 
	 * @return the compute pool names.
	 * @throws UCSServiceException
	 *             in case of an error while communicating with the server.
	 */
	public String[] getPoolNames() throws UCSServiceException {
		final List<ComputePool> queryRes = configResolveChildren("computePool", "org-root", false);

		if (queryRes == null) {
			return new String[0];
		}

		final String[] res = new String[queryRes.size()];
		int i = 0;
		for (final ComputePool obj : queryRes) {
			res[i++] = obj.getName();
		}

		return res;

	}

	/********
	 * Gets service power state.
	 * 
	 * @param profileDn
	 *            profile DN.
	 * @return service power state.
	 * @throws UCSServiceException .
	 */
	public com.gigaspaces.ucs.domain.out.LsPower getPowerStateForProfile(final String profileDn)
			throws UCSServiceException {
		final List<com.gigaspaces.ucs.domain.out.LsPower> powers = configResolveChildren("lsPower", profileDn, false);

		if (powers.size() == 0) {
			return null;
		}

		if (powers.size() != 1) {
			throw new UCSServiceException("Expected to recieve one result for power query on " + profileDn
					+ ", but got " + powers.size());
		}

		return powers.get(0);

	}

	private <T> T getResponseObject(final Class<T> returnType, final HttpResponse response) throws UCSServiceException {
		// Get the HTTP Response
		final HttpEntity entity = response.getEntity();
		if (entity != null) {
			InputStream instream = null;
			try {
				instream = entity.getContent();
			} catch (final IllegalStateException e) {
				throw new UCSServiceException("Failed to read HTTP response: " + e, e);
			} catch (final IOException e) {
				throw new UCSServiceException("Failed to read HTTP response: " + e, e);
			}

			String xmlResult;
			try {
				xmlResult = getResponseXML(instream);
			} catch (final IOException e) {
				throw new UCSServiceException("Failed to read response XML: " + e, e);
			}
			logger.fine("Response: " + xmlResult);

			// Parse the response
			JAXBElement<T> element = null;
			try {
				element = unmarshaller.unmarshal(new StreamSource(new StringReader(xmlResult)), returnType);
			} catch (final JAXBException e) {
				throw new UCSServiceException("Failed to parse response XML: " + e, e);
			}

			return element.getValue();
		}

		return null;
	}

	private String getResponseXML(final InputStream instream) throws IOException {
		final BufferedReader reader = new BufferedReader(new InputStreamReader(instream));
		final StringBuilder builder = new StringBuilder();
		String currentLine = null;
		while (null != (currentLine = reader.readLine())) {
			builder.append(currentLine);
		}
		reader.close();
		final String xmlResult = builder.toString();
		return xmlResult;
	}

	/**********
	 * Gets a service profile DN by its MAC address.
	 * 
	 * @param mac
	 *            the mac address.
	 * @param macPoolDN
	 *            the map pool DN.
	 * 
	 * @return the service profile DN, or null if not found.
	 * @throws UCSServiceException .
	 */
	public String getServerDNFromMAC(final String mac, final String macPoolDN) throws UCSServiceException {
		return getServerDNFromMAC(new String[] { mac }, macPoolDN);
	}

	/**********
	 * Gets a service profile DN by its MAC addresses.
	 * 
	 * @param macs
	 *            the mac addresses.
	 * @param macPoolDN
	 *            the map pool DN.
	 * 
	 * @return the service profile DN, or null if not found.
	 * @throws UCSServiceException .
	 */
	@SuppressWarnings("unchecked")
	public String getServerDNFromMAC(final String[] macs, final String macPoolDN) throws UCSServiceException {
		final Set<String> set = new HashSet<String>(Arrays.asList(macs));

		final List<MacpoolPooled> pooledMacs = getMacPool(macPoolDN);

		logger.info("Found " + pooledMacs.size() + " MACs in pool");

		for (final MacpoolPooled pooledMac : pooledMacs) {
			if (set.contains(pooledMac.getId())) {
				// found you
				final String dn = pooledMac.getAssignedToDn();
				if (!pooledMac.getAssigned().equalsIgnoreCase("yes")) {
					logger.warning("Got query for mac: " + Arrays.toString(macs) + " but mac " + pooledMac.getId()
							+ " is not assigned");
					return null;
				}

				// remove ethernet interface from DN
				final String[] parts = dn.split("/");
				final StringBuilder builder = new StringBuilder();
				for (int i = 0; i < parts.length - 1; i++) {
					builder.append(parts[i]);
					if (i < parts.length - 2) {
						builder.append("/");
					}
				}

				final String result = builder.toString();
				return result;
			}
		}

		logger.warning("Got query for mac: " + Arrays.toString(macs) + " but this mac was not found in mac pool "
				+ macPoolDN);

		return null;
	}

	/******
	 * Get a service profile by its DN.
	 * 
	 * @param dn
	 *            the DN.
	 * @return the profile.
	 * @throws UCSServiceException .
	 */
	public com.gigaspaces.ucs.domain.out.LsServer getServiceProfile(final String dn) throws UCSServiceException {
		final List<com.gigaspaces.ucs.domain.out.LsServer> profiles = getAllServiceProfiles();

		for (final com.gigaspaces.ucs.domain.out.LsServer profile : profiles) {
			if (profile.getDn().equals(dn)) {
				return profile;
			}
		}

		return null;

	}

	/*******
	 * Marshal a domain object into an XML string.
	 * 
	 * @param request
	 *            the domain object.
	 * @return the XML.
	 * @throws UCSServiceException .
	 */
	@SuppressWarnings("unchecked")
	public String getStringFromXMLRequest(final Object request) throws UCSServiceException {
		final StringWriter writer = new StringWriter();
		try {

			marshaller.marshal(new JAXBElement(new QName(getXMLTopLevelName(request)), request.getClass(), request),
					writer);
		} catch (final JAXBException e) {
			throw new UCSServiceException("Failed to marshal request to XML: " + e, e);
		}
		final String content = writer.toString();
		return content;
	}

	public URI getURI() {
		return this.uri;
	}

	private String getXMLTopLevelName(final Object request) {
		final String className = request.getClass().getSimpleName();
		return Character.toLowerCase(className.charAt(0)) + className.substring(1);
	}

	/*******
	 * Checks if a compute pool is empty.
	 * 
	 * @param poolName
	 *            the pool name.
	 * @return true if empty.
	 * @throws UCSServiceException
	 *             in case of an error.
	 */
	public boolean isComputePoolEmpty(final String poolName) throws UCSServiceException {
		final List<ComputePooledSlot> pool = getComputePool(poolName);
		for (final ComputePooledSlot computePooledSlot : pool) {
			if (!computePooledSlot.getAssigned().equalsIgnoreCase("Yes")) {
				return false;
			}
		}
		return true;

	}

	private boolean isLoggedIn(final String username, final String password) {
		if (cookie == null) {
			return false;
		}

		final AaaRefresh refresh = new AaaRefresh();
		refresh.setInCookie(this.cookie);
		refresh.setInName(username);
		refresh.setInPassword(password);

		try {
			final com.gigaspaces.ucs.domain.out.AaaRefresh result = submit(refresh,
					com.gigaspaces.ucs.domain.out.AaaRefresh.class);
			this.cookie = result.getOutCookie();
		} catch (final UCSServiceException e) {
			return false;
		}

		return true;
	}

	/*******
	 * Checks if a service profile is powered on.
	 * 
	 * @param profileDn
	 *            the profile DN.
	 * @return true if powered on.
	 * @throws UCSServiceException
	 *             in case of an error.
	 */
	public boolean isProfilePoweredOn(final String profileDn) throws UCSServiceException {
		final com.gigaspaces.ucs.domain.out.LsPower power = getPowerStateForProfile(profileDn);
		return power.getState().equals("up");
	}

	private void keepAlive() throws UCSServiceException {
		final AaaKeepAlive ka = new AaaKeepAlive();
		ka.setCookie(this.cookie);

		submit(ka, com.gigaspaces.ucs.domain.out.AaaKeepAlive.class);

	}

	/********
	 * Launches a timed task that will send keep alive messages for this
	 * authenticatged session.
	 */
	public void launchSessionKeepAliveTask() {
		// check that login was completed
		if ((this.refreshPeriod == null) || (this.refreshPeriod.longValue() == 0)) {
			throw new IllegalStateException("Keep Alive task should only be launched " + "after a successful login!");
		}

		if (keepAliveTask != null) {
			throw new IllegalStateException("Keep alive task already running!");
		}

		final long delay = this.refreshPeriod / 2;
		// for testing
		// delay = 10;

		keepAliveTask = keepAliveExecutor.scheduleWithFixedDelay(new Runnable() {

			@Override
			public void run() {

				try {
					keepAlive();
				} catch (final UCSServiceException e) {
					logger.severe("Keep Alive failed! Shutting down keep alive task. Error was: " + e);
					shutdownSessionkeepAliveTask();
				}
			}

		}, delay, delay, TimeUnit.SECONDS);

	}

	/*******
	 * Logs in to the UCS manager.
	 * 
	 * @param username
	 *            .
	 * @param password
	 *            .
	 * @throws UCSServiceException
	 *             if authentication failed.
	 */
	public void login(final String username, final String password) throws UCSServiceException {

		if (isLoggedIn(username, password)) {
			return;
		}

		final AaaLogin loginObject = new AaaLogin();

		loginObject.setInName(username);
		loginObject.setInPassword(password);

		final com.gigaspaces.ucs.domain.out.AaaLogin res = submit(loginObject,
				com.gigaspaces.ucs.domain.out.AaaLogin.class);

		this.cookie = res.getOutCookie();
		if ((this.cookie == null) || (this.cookie.length() == 0)) {
			throw new UCSServiceException("Failed to login with username/password: " + username + "/" + password
					+ ": received null login");
		}
		this.refreshPeriod = res.getOutRefreshPeriod();

	}

	private String normalizeOrgName(final String orgName) {
		final StringBuilder builder = new StringBuilder();
		final String[] parts = orgName.split("/");

		for (int i = 0; i < parts.length; i++) {

			final String part = parts[i];
			if (!part.startsWith("org-")) {
				builder.append("org-");
			}
			builder.append(part);
			if (i < parts.length - 1) {
				builder.append("/");
			}

		}
		return builder.toString();

	}

	private void setFastBoot(final boolean val) {
		System.setProperty("com.sun.xml.internal.bind.v2.runtime.JAXBContextImpl.fastBoot", Boolean.toString(val));
	}

	/********
	 * Shut down a server.
	 * 
	 * @param dn
	 *            the service DN.
	 * @param softShutDown
	 *            true if a soft (OS) shut down is performed, false if a hard
	 *            (power) shut down.
	 * @return the modified power state.
	 * @throws UCSServiceException .
	 */
	public com.gigaspaces.ucs.domain.out.LsPower shutdownServer(final String dn, final boolean softShutDown)
			throws UCSServiceException {

		final String state = (softShutDown ? "soft-shut-down" : "down");
		return applyPowerState(dn, state);
	}

	/*****
	 * Stops the session keep alive timer.
	 */
	public void shutdownSessionkeepAliveTask() {
		keepAliveTask.cancel(false);
		keepAliveTask = null;
	}

	/***********
	 * Creates a service profile from a template.
	 * 
	 * @param templateName
	 *            the service profile template name.
	 * @param targetOrg
	 *            the organization where the service profile will be created.
	 * @param serverName
	 *            the name of the new profile.
	 * @return the created profile.
	 * @throws UCSServiceException .
	 */
	public com.gigaspaces.ucs.domain.out.LsServer startServerFromTemplate(final String templateName,
			final String targetOrg, final String serverName) throws UCSServiceException {

		final LsInstantiateTemplate inst = new LsInstantiateTemplate();
		inst.setCookie(this.cookie);
		inst.setDn(targetOrg + "/ls-" + templateName);
		inst.setInTargetOrg(targetOrg);
		inst.setInServerName(serverName);
		inst.setInHierarchical("false");

		final com.gigaspaces.ucs.domain.out.LsInstantiateTemplate res = submit(inst,
				com.gigaspaces.ucs.domain.out.LsInstantiateTemplate.class);

		final com.gigaspaces.ucs.domain.out.LsServer server = 
			(com.gigaspaces.ucs.domain.out.LsServer) getElementFromJAXBList(
				res.getContent(), ConfigConfig.class).getManagedObject().getValue();

		return server;

	}

	private <T> T submit(final Object request, final Class<T> returnType) throws UCSServiceException {
		// First marshal the request object using the JAXB Annotations
		final String content = getStringFromXMLRequest(request);

		return submitXML(content, returnType);

	}

	private <T> T submitXML(final String content, final Class<T> returnType) throws UCSServiceException {

		// update the last server access time. Useful for keep alive tests.
		this.lastAccessed = new Date();

		logger.fine("Request: " + content);
		// Create the post request
		final HttpPost post = new HttpPost(uri);
		try {
			post.setEntity(new StringEntity(content));
		} catch (final UnsupportedEncodingException e) {
			throw new UCSServiceException("Failed to create HTTP request: " + e, e);
		}

		// Execute the Post request
		HttpResponse response;
		try {
			response = httpClient.execute(post);
		} catch (final IOException e) {
			throw new UCSServiceException("Failed to send HTTP request: " + e, e);
		}

		final T retObj = getResponseObject(returnType, response);

		checkForErrorResponse(retObj);
		return retObj;

	}

	/********
	 * Creates a domain object from the given XML string.
	 * 
	 * @param xml
	 *            the xml string.
	 * @return the domain object.
	 * @throws UCSServiceException .
	 */
	public Object unmarshalEvent(final String xml) throws UCSServiceException {
		try {
			final JAXBElement<?> element = (JAXBElement<?>) eventUnmarshaller.unmarshal(new StringReader(xml));
			if (element != null) {
				final Object event = element.getValue();
				return event;
			} else {
				return null;
			}
		} catch (final JAXBException e) {
			throw new UCSServiceException("Failed to unmarshal event XML", e);
		}
	}

	private DefaultHttpClient useTrustingTrustManager(final DefaultHttpClient httpClient) {
		try {
			// First create a trust manager that won't care.
			final X509TrustManager trustManager = new X509TrustManager() {
				public void checkClientTrusted(final X509Certificate[] chain, final String authType)
						throws CertificateException {
					// Don't do anything.
				}

				public void checkServerTrusted(final X509Certificate[] chain, final String authType)
						throws CertificateException {
					// Don't do anything.
				}

				public X509Certificate[] getAcceptedIssuers() {
					// Don't do anything.
					return null;
				}
			};

			// Now put the trust manager into an SSLContext.
			final SSLContext sslcontext = SSLContext.getInstance("TLS");
			sslcontext.init(null, new TrustManager[] { trustManager }, null);

			// Use the above SSLContext to create your socket factory
			// (I found trying to extend the factory a bit difficult due to a
			// call to createSocket with no arguments, a method which doesn't
			// exist anywhere I can find, but hey-ho).
			final SSLSocketFactory sf = new SSLSocketFactory(sslcontext);
			sf.setHostnameVerifier(SSLSocketFactory.ALLOW_ALL_HOSTNAME_VERIFIER);

			// If you want a thread safe client, use the ThreadSafeConManager,
			// but
			// otherwise just grab the one from the current client, and get hold
			// of its
			// schema registry. THIS IS THE KEY THING.
			final ClientConnectionManager ccm = httpClient.getConnectionManager();
			final SchemeRegistry schemeRegistry = ccm.getSchemeRegistry();

			// Register our new socket factory with the typical SSL port and the
			// correct protocol name.
			schemeRegistry.register(new Scheme("https", sf, HTTPS_PORT));

			// Finally, apply the ClientConnectionManager to the Http Client
			// or, as in this example, create a new one.
			return new DefaultHttpClient(ccm, httpClient.getParams());
		} catch (final Throwable t) {
			// AND NEVER EVER EVER DO THIS, IT IS LAZY AND ALMOST ALWAYS WRONG!
			t.printStackTrace();
			return null;
		}
	}

	@SuppressWarnings("unchecked")
	private <T> JAXBElement<T> wrapObjectInJAXBElement(final T obj) {
		return new JAXBElement<T>(new QName(getXMLTopLevelName(obj)), (Class<T>) obj.getClass(), obj);
	}

}
